/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture;

import picture.gui.*;
import java.io.*;
import javax.swing.JFrame;

/**
 * Eine Transformation bildet ein Layer auf ein anderes Layer ab. Beispiel:
 * Die Bestimmung der Fouriertransformierten des Pictures <tt>p</tt> koennte mit
 * Hilfe einer Unterklasse <tt>FFT</tt> etwa wie folgt realisiert werden:
 * <pre>
 *  Layer fftVonP = (new FFT()).of(p);
 * </pre>
 * Etwaige Parameter koennten z.B. im Konstruktor uebergeben werden. 
 * Haengt eine Transformation von zwei (oder mehr) Bildern ab, 
 * so ist eine Realisierung in folgender Gestalt denkbar:
 * <pre>
 *  Layer a, b;
 *  Layer aSternB = (new Falter(a)).of(b);
 * </pre>
 *
 * DefaultTransformation ist die Standard-Implementation von Transformation.
 */
public abstract class DefaultTransformation implements ProgressViewableTransformation {

    private static int filterNummer = 0;

    /**
     * Bestimmt den mittleren quadratischen Abstand zweier Bilder.
     * Dieser Abstand wird fuer Real- und Imaginaerteil getrennt berechnet.
     * Zurueckgegeben wird eine komplexe Zahl, bei der im Realteil das
     * Ergebnis der reellen Bilder und im Imaginaerteil das Ergebnis der
     * imaginaeren Bilder sitzt.
     * @param p: das eine Bild
     * @param q: das andere Bild
     * @return mittlerer quadratischer Abstand zwischen <tt>p</tt> und <tt>q</tt>.
     * @exception PictureException, falls die Bilder <tt>p</tt> und <tt>q</tt> nicht 
     * die gleiche Groesse haben.
     */
    public static final double getDistance(Picture p, Picture q) 
	throws PictureException {

	int w = p.getWidth(); int h = p.getHeight();
	if (w != q.getWidth() || h != q.getHeight() || 
	    p.getNumberOfLayers() != q.getNumberOfLayers())
	    throw new PictureException("Die Bilder haben unterschiedliche Gren.");

	double d = 0;
	double e;

	for (int z = 0; z < p.getNumberOfLayers(); z++)
	    for (int x = 0; x < w; x++)
		for (int y = 0; y < h; y++) {
		    e = p.getLayer(z).getValue(x, y) - 
			q.getLayer(z).getValue(x, y);
		    d += sqr(e);
		}

	return d / (w * h * p.getNumberOfLayers());
    }

    /**
     * Hilfsmethode. Quadriert das Argument
     */
    private static final double sqr(double x) {return x * x;}

    /**
     * Prueft, ob die uebergebene Zahl eine Potenz von zwei ist.
     * @param i die zu pruefende Zahl
     * @return <tt>true</tt>, falls es ein ganzes <i>n</i> gibt mit <i>i</i>=2<sup>n</sup>
     */
    public static final boolean isPowerOfTwo(int i) {

	if (i == 0) return false;
	if (i < 0) i = -i;
	
	for (; i > 1; i /= 2)
	    if (i % 2 == 1) return false;

	return true;
    }

    /**
     * Erzeugt ein Layer von der selben Instanz wie <tt>p</tt> mit dem Inhalt von
     * <tt>data</tt>
     * @param p: Vorlage
     * @param data: Inhalt des neuen Bildes
     * @return ein Layer wie <tt>p</tt>, mit <tt>data</tt> als Inhalt.
     */
    protected Layer createNewInstance(Layer p, double[][] data) {

	return p.createNewInstance("", data);
    }

    /**
     * Liefert null zurck. Default fr keinen Optionendialog.
     */
    public DialogPane filterOptionPane(JFrame owner, Picture pic) 
	throws PictureException {
	return null;
    }

    /**
     * Speichert diese Transformation in der angegebenen Datei. Klassen, die von
     * DefaultTransformation erben, m&uuml;ssen beruecksichtigen, dass alle
     * Referenz-Instanzvariablen entwender java.io.Seriazable implementieren oder
     * transient deklariert sind, d.h. nicht mit abgespeichert werden.
     * @param file: Zieldatei
     * @exception IOException, falls ein Fehler auftritt.
     */
    public void saveTo(File f) throws IOException {

	ObjectOutputStream ous = new ObjectOutputStream(new FileOutputStream(f));
	ous.writeObject(this);
	ous.close();
    }

    /**
     * Laedt einen Filter aus einer Datei
     */
    public static Transformation loadFrom(File f) throws IOException, ClassNotFoundException {

	ObjectInputStream in = new ObjectInputStream(new FileInputStream(f));
	Transformation t = (Transformation)in.readObject();
	in.close();
	return t;
    }

    /**
     * Fuehrt die Transformation auf einer Bildebene aus. 
     */
    public abstract Layer of(ProgressHandler ph, Layer p) throws PictureException;

    /**
     * Fuehrt die Transformation Ebenenweise auf dem uebergebenen Bild aus.
     */
    public Picture of(Picture p) throws PictureException {

	int z = p.getNumberOfLayers();	
	Layer[] pics = new Layer[z];

	for (int i = 0; i < z; i++) 
	    pics[i] = of(p.getProgressHandler(), p.getLayer(i));

	return new Picture(p, pics);
    }

    /**
     * Gibt einen Dateinamen zurueck, der beim Speicher zunaechst vorgeschlagen wird.
     * Diese Implementierung liefert den String "filter*.jf", wobei * von 1 beginnend
     * bei jedem Methodenaufruf eins hochzaehlt. Unterklassen moegen diese Methode
     * ueberschreiben, indem sie z.B. "filter" durch spezifischere Bezeichnungen
     * ersetzen. Von einer Aenderung der Dateiendung ".jf" sollte hingegen abgesehen werden.
     *
     * @return automatisch generierter Vorschlag fuer einen Dateinamen
     */
    public String getPreferredFileName() {

	return "filter" + (++filterNummer) + ".jf";
    }
}

