/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture;

import java.awt.*;
import javax.swing.event.*;
import java.util.*;
import java.io.*;

import picture.filetype.*;

/**
 * Ein Picture ist ein Layer mit mehreren Ebenen. Jede Ebene ist ein
 * Layer. Ebenen koennen z.B. Real- und Imaginaerteil eines komplexwertigen 
 * Bildes oder Rot-, Gruen- und Blau-Anteil eines Farbbildes sein. Zur Generierung
 * eines Image-Objektes dient eine Instanz der Klasse Presentation. 
 */
public class Picture implements PictureModel, PictureConstants, PictureProperty {

    // #############################################################################
    // ###  Instanzvariablen  ######################################################
    // #############################################################################

    protected Layer[] layers;
    protected int width;
    protected int height;
    protected Presentation presentation = PresentationFactory.createDefaultPresentation(this);
    protected PictureType pictureType = PictureTypeFactory.getDefaultPictureType();
    
    protected transient String fileName = "";
    protected transient ProgressHandler ph;
    protected transient Picture preview;
    
    private transient Vector changeListeners = new Vector();

    /**
     * Konvertiert ein Bild in ein anderes PictureType.
     *
     * @param p Bild, das ins andere Format umgewandelt werden soll
     * @param pt PictureType des gewnschten Formats
     */
    public Picture convert(Picture p, PictureType pt) throws PictureException {
	p.pictureType = pt;
	return p;
    }

    // #############################################################################
    // ###  Konstruktoren  #########################################################
    // #############################################################################

    /**
     * Erzeugt ein neues Picture mit einem Layer
     */
    public Picture() {

	layers = new Layer[]{new Layer()};
	width = layers[0].getWidth();
	height = layers[0].getHeight();
    }

    /**
     * Erzeugt ein neues Picture mit den angegebenen Layern. 
     *
     * @param layers: Ebenen des zu erzeugenden Bildes
     */
    public Picture(Layer[] layers) {

	this.layers = layers;
	width = layers[0].getWidth();
	height = layers[0].getHeight();
    }

    /**
     * Wie <tt>MulitLayerPicture(Layer[] layers)</tt>, nur da&szlig; zus&auml;tzlich
     * der Dateiname festgelegt wird.
     *
     * @param layers: Ebenen des zu erzeugenden Bildes
     */
    public Picture(Layer[] layers, String fileName) {

	this(layers);
	this.fileName = fileName;
    }

    /**
     * Erzeugt ein neues Picture.
     * Es werden die uebergebenen layer gesetzt und alle anderen Daten vom uebergebenen
     * Picture uebernommen. (Ausnahmen: <tt>preview = null, fileName = ""</tt>)
     * Dieser Konstruktor soll die Funktion des frueheren <tt>createNewInstance</tt>
     * aus der Layer-Klasse uebernehmen.
     *
     * @param source: Quell-Picture
     * @param layers: Ebenen des Bildes
     */
    public Picture(Picture source, Layer[] layers) {

	this(source, layers, "");
    }

    /**
     * Wie <tt>Picture(Picture, Layer[])</tt>, nur da&szlig; 
     * zus&auml;tzlich der Dateiname festgelegt wird.
     *
     * @param source: Quell-Picture
     * @param layers: Ebenen des Bildes
     * @param fileName: Dateiname des Bildes
     */
    public Picture(Picture source, Layer[] layers, String fileName) {

	this(layers, fileName);

	if (source == null) return;

	getToolsFrom(source);
    }

    // #############################################################################
    // ###  EventListener-Mechanismus  #############################################
    // #############################################################################

    /**
     * Registriert einen ChangeListener.
     * ChangeEvents werden ausgefuehrt, sobald sich die Ansicht des Bildes 
     * veraendert, z.B. durch Auswahl einer anderen Ebene oder durch Veraederung
     * der Farbbasis etc. Unterklassen haben die M&ouml;glichkeit, 
     * mit Hilfe der Methode <tt>fireChangeEvent()</tt> ChangeEvents auszul&ouml;sen.
     * 
     * @param cl: neu hinzuzufuegender ChangeListener
     * @see #fireChangeEvent
     * @see #removeChangeListener
     */
    public void addChangeListener(ChangeListener cl) {changeListeners.addElement(cl);}

    /**
     * L&ouml;scht einen registrierten ChangeListener.
     * 
     * @param cl: der zu loeschende ChangeListener
     * @see #addChangeListener
     */
    public void removeChangeListener(ChangeListener cl) {changeListeners.removeElement(cl);}

    /**
     * Erzeugt ein ChangeEvent und f&uuml;rt dieses auf allen registrierten 
     * ChangeListenern aus. Per Konvention werden ChangeEvents ausgel&ouml;st, 
     * wenn sich an der Ansicht des Bildes etwas &aouml;ndert, etwa durch Auswahl
     * einer anderen Darstellungsebene.
     * Zur Ausf&uuml;hrung des ChangeEvents erzeugt die Methode <i>keinen</i> eigenen
     * Thread, so da&szlig; bei flei&szlig;igen ChangeListenern damit zu rechnen ist,
     * da&szlig; der Thread, der den Aufruf dieser Methode verursacht, ausgebremst
     * wird.
     * Zur Verwaltung der ChangeListener stehen die Methoden 
     * <tt>addChangeListener(ChangeListener)</tt> und
     * <tt>removeChangeListener(ChangeListener)</tt> zur Verf&uuml;gung.
     *
     * @see #addChangeListener
     * @see #removeChangeListener
     */
    protected void fireChangeEvent() {

	ChangeEvent ce = new ChangeEvent(this);
	Enumeration e = changeListeners.elements();
	while (e.hasMoreElements())
	    ((ChangeListener)e.nextElement()).stateChanged(ce);
    }

    /**
     * Setzt einen ProgressHandler
     * <p>
     * Einige Methoden dieser Klasse unterstuetzen den folgenden ProgressHandler-Mechanismus:
     * Von Zeit zu Zeit werfen sie ein ProgressEvent aus, das von einem ProgressHandler
     * abgehoert werden kann, der seinerseits dann z.B. die enthaltenen Daten an eine
     * Fortschrittsanzeige (z.B. JProgressBar) weiterleiten kann. 
     * @see ProgressHandler
     * @see #getProgressHandler
     * @param ph der neu zu setzende ProgressHandler 
     */
    public void setProgressHandler(ProgressHandler ph) {this.ph = ph;}

    /**
     * Fragt den gesetzten ProgressHandler ab. Falls kein ProgressHandler mittels
     * <tt>setProgressHandler</tt> angegeben wurde, wird <tt>null</tt> zurueckgegeben.
     * @see #setProgressHandler
     * @return der aktuelle ProgressHandler
     */
    public ProgressHandler getProgressHandler() {return ph;}

    /**
     * Schiesst ein ProgressEvent ab. Falls eine Methode diesen Mechanismus unterstuezten
     * will, so muss sie folgendes tun: 
     * Im Kommentar ist die genaue Anzahl der voraussichtlich abgefeuerten Events anzugeben,
     * in Abhaengigkeit von von aussen sichtbaren Groessen. Innerhalb der Methode kann dann
     * entsprechend dieser Anzahl moeglichst in etwa gleichen Abstaenden ein progressEvent
     * abgefeuert werden, das z.B. einer Fortschrittsanzeige die Erhoehung um eins mitteilt, 
     */
    public void fireProgressEvent() {if (ph != null) ph.step();}

    // #############################################################################
    // ###  Implementierung von PictureModel  ######################################
    // #############################################################################

    /**
     * Erzeugt ein <tt>java.awt.Image</tt>-Objekt, das die aktuelle Ansicht dieses
     * Bildes zeigt. 
     * Wenn sich das Ergebnis dieser Methode aendert, wird ein ChangeEvent auf den
     * registrierten ChangeListenern ausgef&uuml;hrt.
     * Die Methode dient zur Implementierung von <tt>PictureModel</tt>.
     *
     * @see #fireChangeEvent
     * @see PictureModel
     * @return aktuelle Ansicht dieses Bildes.
     */
    public Image getImage() {
	
	try {
	    return presentation.getImage();
	} catch (PictureException e) {} // temporaere Inkompatibilitaet.

	try {
	    return PresentationFactory.createDefaultPresentation(this).getImage();
	} catch (PictureException neveBeenThrown) {}
	return null;
    }

    /**
     * Gibt die Breite dieses Bildes zur&uuml;ck. Die Methode dient zur Implementierung
     * von <tt>PictureModel</tt>. Der zur&uuml;ckgegebene Wert dieser Methode 
     * &auml;ndert sich w&auml;hrend der Lebensdauer des Objekts nicht.
     *
     * @see #getHeight
     * @see PictureModel
     * @return Breite dieses Bildes
     */
    public int getWidth() {return width;}

    /**
     * Gibt die H&ouml;he dieses Bildes zur&uuml;ck. Die Methode dient zur Implementierung
     * von <tt>PictureModel</tt>. Der zur&uuml;ckgegebene Wert dieser Methode 
     * &auml;ndert sich w&auml;hrend der Lebensdauer des Objekts nicht.
     *
     * @see #getWidth
     * @see PictureModel
     * @return H&ouml;he dieses Bildes
     */
    public int getHeight() {return height;}

    /**
     * liefert den PictureType zur&uuml;ck, dem dieses Bild angeh&ouml;rt. 
     * Falls das Bild nicht von einer Unterklasse erzeugt wurde und bei der
     * Konstruktion kein QuellBild angegeben wurde, also kein Typ bekannt ist,
     * so liefert diese Methode <tt>null</tt> zur&uuml;ck.
     * 
     * @return Der Typ dieses Bildes.
     */
    public PictureType getPictureType() {return pictureType;}

    /**
     * Setz den PictureType dieses Bildes.
     *
     * @param pt neuer PictureType
     */
    public void setPictureType(PictureType pt) {
	pictureType = pt;
    }

    /**
     * Erzeugt eine Zeichenkette mit Informationen zum Bild an der Position 
     * (<i>x</i>, <i>y</i>). Falls diese Position kein g&uuml;ltiger Bildpunkt ist,
     * wird die leere Zeichenkette zur&uuml;ckgegeben.
     *
     * @param x: Abszisse 
     * @param y: Ordinate
     * @return Informationen zum Bild an der Position (<i>x</i>, <i>y</i>).
     */
    public String getInfo(int x, int y) {return "";}

    // #############################################################################
    // ###  nicht-statische Methoden zu Lesen und schreibenn  ######################
    // #############################################################################
    
    /**
     * Gibt den Dateinamen zur&uuml;ck, falls dieser bekannt ist.
     *
     * @return Dateiname dieses Bildes.
     */
    public String getFileName() {return fileName;}

    /**
     * Setzt den Dateinamen dieses Bildes.
     *
     * @param fileName neuer Dateiname dieses Bildes.
     */
    public void setFileName(String fileName) {this.fileName = fileName;}

    /**
     * Speichert diese Datei. Als PictureProperty wird dieses Objekt
     * verwendet.  Die Methode ruft intern
     * <tt>save(OutputStream,PictureProperty)</tt> des verwendeten
     * PictureTypes auf.
     * 
     * @param fileName Name der Datei
     * @see #save(OutputStream,PictureProperty)
     * @exception IOException, falls beim Schreiben ein Fehler auftritt
     * @exception PictureException, falls z.B. irgendwelche Daten fehlen.  */
    protected void saveToFile(String fileName) throws IOException, PictureException {
	saveToFile(fileName, this);
    }

    /**
     * Speichert diese Datei. Als PictureProperty wird dieses Objekt
     * verwendet. Als Dateiname wird <tt>getFileName()</tt> verwendet.
     *
     * @see #getFileName
     * @see #save(OutputStream,PictureProperty)
     */
    public void saveToFile() throws IOException, PictureException {
	saveToFile(fileName, this);
    }

    /**
     * Speichert diese Datei. Als Dateiname wird <tt>getFileName()</tt> verwendet.
     * Die Methode ruft intern <tt>save(OutputStream,PictureProperty)</tt> auf.
     *
     * @see #getFileName
     * @see #save(OutputStream,PictureProperty)
     * @param pp PictureProperty-Objekt mit den Daten zum Dateiformat.
     */
    public void saveToFile(PictureProperty pp) throws IOException, PictureException {
	saveToFile(fileName, pp);
    }

    /**
     * Speichert diese Datei. 
     * Diese Methode ruft intern <tt>save(OutputStream,PictureProperty)</tt> auf.
     *
     * @param fileName Dateiname
     * @see #save(OutputStream,PictureProperty)
     * @param pp PictureProperty-Objekt mit den Daten zum Dateiformat
     */
    public void saveToFile(String fileName, PictureProperty pp) throws IOException, PictureException {
	this.fileName = fileName;
	save(new FileOutputStream(fileName), pp);
    }

    /**
     * Speichert diese Datei.
     * 
     * @param out OutputStream, in den gespeichert werden soll.
     * @param pp PictureProperty-Objekt mit den Daten zum Dateiformat
     */
    public void save(OutputStream out, PictureProperty pp) throws IOException, PictureException {
	if(pictureType != null)
	    pictureType.save(this, pp, out);
	else
	    throw new IOException("Kein Dateiformat spezifiziert.");
    }

    /**
     * &Uuml;bernimmt alle Daten aus dem angegebenen Bild <tt>p</tt>.
     *
     * @param p Quellbild.
     */
    protected void getDataFrom(Picture p) {

	layers = p.layers;
	width = p.getWidth();
	height = p.getHeight();
	preview = p.preview;

	getToolsFrom(p);
    }

    /**
     * Uebertraegt Presentation, PictureType, ProgressHandler und
     * die ChangeListener von source auf this.
     */
    private void getToolsFrom(Picture source) {

	presentation = source.getPresentation().getCopy(this);
	pictureType = source.getPictureType();

	ph = source.getProgressHandler();
	Enumeration e = source.changeListeners.elements();
	while (e.hasMoreElements())
	    changeListeners.addElement(e.nextElement());
    }

    // #############################################################################
    // ###  Verschiedenes  #########################################################
    // #############################################################################

    /**
     * Legt die Praesentation dieses Bildes fest. Mit Hilfe des durch diese Methode
     * spezifizierten Presentation-Objekts wird festgelegt, wie die Methode 
     * <tt>getImage()</tt> aus den verschiedenen Ebenen dieses Bildes eine
     * Ansicht erzeugt.
     *
     * @param p das neuen Presentationsobjekt.
     * @see #getImage
     * @see #getPresentation
     * @see Presentation
     */
    public void setPresentation(Presentation p) {presentation = p;}

    /**
     * Setzt die Presentation dieses Bildes auf die StandardPraesentation
     */
    public void setDefaultPresentation() {

	presentation = PresentationFactory.createDefaultPresentation(this);
    }

    /**
     * Gibt die aktuelle Praesentation zurueck.
     *
     * @see #setPresentation
     * @return aktuelle Praesentatioin
     */
    public Presentation getPresentation() {return presentation;}

    /**
     * Gibt die Anzahl der Layer dieses Bildes zur&uuml;ck.
     * 
     * @return Anzahl der Layer dieses Bildes.
     */
    public int getNumberOfLayers() {return layers.length;}

    /**
     * Gibt den Layer mit der angegebenen Nummer zur&uuml;ck. Im Fall
     * <tt>i < 0</tt> oder <tt>i >= getNumberOfLayers()</tt> wird <tt>null</tt>
     * zur&uuml;ckgegeben.
     *
     * @param i: Index des gew&uuml;nschten Layers.
     * @return <i>i</i>-ter Layer.
     */
    public Layer getLayer(int i) {

	if (i < 0 || i >= getNumberOfLayers())
	    return null;

	return layers[i];
    }

    /**
     * Erzeugt ein Bild, das durch zyklische Verschiebung dieses Bildes in 
     * Hoehe und Breite entsteht.
     *
     * @param width Verschiebung in <tt>x</tt>-Richtung
     * @param height Verschiebung in <tt>y</tt>-Richtung
     */
    public Picture getShiftedInstance(int width, int height) {

	Layer[] p = new Layer[getNumberOfLayers()];

	for (int i = 0; i < layers.length; i++)
	    p[i] = layers[i].getShiftedInstance(width, height);

	return new Picture(this, p);
    }

    /**
     * gibt das kleinste quadratische Bild zurueck mit Kantenlaenge 2^<i>n</i>,
     * in das dieses Bild hineinpasst. Dieses Bild wird in dem neu konstruierten
     * Bild zentriert dargestellt. Die Raender werden mit 0 aufgefuellt.
     *
     * @return naechstgroesseres quadratisches Zweierpotenzbild
     */
    public Picture getEmbeddedInstance() {

	Layer[] p = new Layer[getNumberOfLayers()];

	for (int i = 0; i < layers.length; i++)
	    p[i] = layers[i].getEmbeddedInstance();

	return new Picture(this, p);
    }

    /**
     * Erzeugt eine in der Groesse skalierte Version dieses Bildes. 
     * @param width Breite des neuen Bildes
     * @param height Hoehe des neuen Bildes
     * @return eine Projektion dieses Bildes auf die Groesse <tt>width</tt> mal <tt>height</tt>.
     */
    public Picture getScaledInstance(int width, int height) {

	Layer[] p = new Layer[getNumberOfLayers()];

	for (int i = 0; i < layers.length; i++)
	    p[i] = layers[i].getScaledInstance(width, height);

	return new Picture(this, p);
    }

    /**
     * Erzeugt eine Version dieses Pictures, in dem Layer <i>i</i> mit dem
     * <tt>d[</tt><i>i</i><tt>]</tt> gewichtet ist. Insbesondere mu&szlig; die
     * L&auml;nge des &uuml;bergebenen Feldes also gleich der Anzahl der Ebenen
     * dieses Bildes sein.
     *
     * @param d Vektor mit Gewichten
     * @return mit <tt>d</tt> gewichtete Version dieses Pictures
     */
    public Picture getWeightedInstance(double[] d) {

	Layer[] p = new Layer[getNumberOfLayers()];

	for (int i = 0; i < layers.length; i++)
	    p[i] = layers[i].getWeightedInstance(d[i]);

	return new Picture(this, p);
    }

    /**
     * Erzeugt ein Vorschaubild aus diesem Objekt. Fuer Dialogfenster kann diese 
     * Methode zur Erzeugung einer Kopie dieses Pictures in der Standardgroesse 
     * 64 mal 64 verwendet werden. 
     * @return Ein neues Layer, das dieses Bild in der Normgroesse 64 mal 64 enthaelt.
     */
    public Picture getPreview() {

	if (preview != null) return preview;

	Layer [] p = new Layer[getNumberOfLayers()];

	for (int i = 0; i < layers.length; i++)
	    p[i] = layers[i].getPreview();

	Picture mlp = new Picture(this, p);
	return mlp;
    }

    /**
     * Gibt den Wert dieses Bildes an der Position (<i>x</i>, <i>y</i>) an.
     * Falls die angegebene Bildposition nicht gueltig ist, wird <tt>null</tt>
     * zurueckgegeben.
     * 
     * @param x Abszisse
     * @param y Ordinate
     * @return Wert am Punkt (<i>x</i>, <i>y</i>).
     */
    public double[] getValue(int x, int y) {

	if (x < 0 || y < 0 || x >= width || y >= height)
	    return null;

	double[] value = new double[layers.length];

	for (int i = 0; i < layers.length; i++)
	    value[i] = layers[i].getValue(x, y);

	return value;
    }

    /**
     * Gibt an, ob dieses Bild semantisch gesehen ein Farbbild ist, d.h. ob z.B. die
     * verschiedenen Layer f&uuml;r verschiedene Farben stehen.
     * Die Standardimplementierung dieser Methode liefert grunds&auml;tzlich <tt>false</tt>
     * zur&uuml;ck. Unterklassen m&ouml;gen sie geeignet &uuml;berschreiben.
     *
     * @return ist dieses Bild ein Farbbild?
     */
    public boolean isColored() {
	return getNumberOfLayers() == 3;
    }

    /**
     * Gibt an, ob dieses Bild semantisch gesehen ein komplexwertiges Bidl ist, d.h. ob
     * es zwei Ebenen hat und die eine Ebene als Real- und die andere Ebene als Imagin&auml;rteil
     * aufgefa&szlig;t werden kann.
     * Die Standardimplementierung dieser Methode liefert grunds&auml;tzlich <tt>false</tt>
     * zur&uuml;ck. Unterklassen m&ouml;gen sie geeignet &uuml;berschreiben.
     *
     * @return ist dieses Bild komplexwertig?
     */
    public boolean isComplex() {
	return getNumberOfLayers() == 2;
    }

    /**
     * Erzeugt ein Histogramm. Das Feld ergibt sich aus der Summe der Felder
     * <tt>getLayer(k).histogram(i)</tt> (<i>k</i>=1,...,getNumberOfLayers())
     */
    public int[] histogram(int i) {

	int[] erg = new int[i];

	for (int k = 0; k < getNumberOfLayers(); k++) {

	    int[] l = getLayer(k).histogram(i);

	    for (int n = 0; n < i; n++)
		erg[n] += l[n];
	}

	return erg;
    }
}
