/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;

/**
 * Eine Presentation dient zur Umwandlung eines Pictures in eine Darstellung, die AWT
 * auf den Bildschirm malen kann, d.h. ein Image.
 *
 * Diese Klasse dient zur Erzeugung und Verwaltung von Presentations. Sie enthaelt nur
 * statische Methoden und ist nicht zum Initialisieren gedacht.
 * 
 * Naehere Informationen zum Presentation-Mechanismus sind im Interface-Kopf von 
 * Presentation zu finden.
 */
public class PresentationFactory {

    private static Vector pres = new Vector();
    private static PresentationType defaultPT = new DefaultPresentation();

    private static final class MenuActionListener implements ActionListener {

	private final Picture p;
	private final PresentationType pt;

	public MenuActionListener(Picture p, PresentationType pt) {

	    this.p = p; this.pt = pt;
	}

	public void actionPerformed(ActionEvent e) {

	    try {
		p.setPresentation(pt.createPresentation(p));
	    } catch (Exception ex) {
		p.setDefaultPresentation();
	    }
	}
    }

    static {
	pres.addElement(defaultPT);
	pres.addElement(new BlackWhitePresentation());
	pres.addElement(new ColorPresentation());
	pres.addElement(new ComplexPresentation());
	pres.addElement(new SelectPresentation());
    }

    /**
     * F&uuml;gt den angegebenen PresentationType in die Liste ein.
     */
    public static void registerPresentationType(PresentationType p) {pres.addElement(p);}

    /**
     * Entfernt den angegebenen PresentationType aus der Liste.
     */
    public static void removePresentationType(PresentationType p) {pres.removeElement(p);}

    /**
     * Gibt eine Enumeration aus, die alle registrierten PictureTypes aufz&auml;hlt.
     */
    public static Enumeration getAllRegisteredPresentationTypes() {return pres.elements();}

    /**
     * Erzeugt eine StandardPresentation auf dem angegebenen Bild
     */
    public static Presentation createDefaultPresentation(Picture p) {

	try {
	    return defaultPT.createPresentation(p);
	} catch (PictureException wirdNieAufgerufen) {};
	return null;
    }    

    /**
     * Sucht in der Liste einen PresentationType, der sich so nennt wie angegeben und
     * erzeugt davon eine Presentation auf dem angegeben Picture. Dem Namen kann ein
     * $, gefolgt von spezifischen Optionen fuer die Presentation angefuegt sein.
     */
    public static Presentation presentationByName(Picture p, String name) 
	throws PictureException {

	Enumeration e = pres.elements();
	while (e.hasMoreElements()) {

	    PresentationType t = (PresentationType) e.nextElement();
	    String tname = t.getName();

	    if (name.startsWith(tname) && t.isCompatible(p)) 
		try {
		    return t.createPresentation(p, name.substring(tname.length() + 1));
		} catch (Exception ex) {
		    return t.createPresentation(p);
		}
	}

	return createDefaultPresentation(p);
    }

    /**
     * Erzeugt ein JPopupMenu, das JRadioButtonMenuItems zu jedem registrierten 
     * PresentationType enth&auml;lt. Passt ein Typ nicht zum angegebenen Picture,
     * wird er aufgef&uuml;hrt, aber auf enabled=false gesetzt. Das Ausw&auml;hlen
     * eines Menueintrags bewirkt, da&szlig; eine Instanz des ausgew&auml;hlten
     * Types erzeugt und auf das Picture angesetzt wird. Zur Information anderer
     * Objekte wird zus&auml;zlich der beigef&uuml;gte ActionListener registriert.
     * Nach Konvention verwenden alle Presentations in ihrer getJPopupMenu-Methode
     * zun&auml;chst die Eintr&auml;ge, die diese Methode liefert, erweitert um
     * m&ouml;gliche spezielle Presentation-Optionen. Um ein Menu zum Anzeigen zu
     * erhalten, wird daher die Verwendung von 
     * <pre>
     * p.getPresentation().getJPopupMenu()
     * </pre>
     * empfohlen.
     */
    public static JPopupMenu getJPopupMenu(Picture p, ActionListener al) {

	JPopupMenu pm = new JPopupMenu("");
	Enumeration e = pres.elements();
	PresentationType ptt = p.getPresentation().getPresentationType();
	PresentationType pt;
	JRadioButtonMenuItem mi;
	ButtonGroup b = new ButtonGroup();
	
	while (e.hasMoreElements()) {

	    pt = (PresentationType) e.nextElement();

	    mi = new JRadioButtonMenuItem(pt.getName());
	    b.add(mi);

	    if (pt.equals(ptt)) 
		mi.setSelected(true);

	    if (!pt.isCompatible(p))
		mi.setEnabled(false);
	    else {
		mi.addActionListener(al);
		mi.addActionListener(new MenuActionListener(p, pt));
	    }
	    
	    pm.add(mi);
	}

	return pm;
    }
}
