/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.filetype;

import javax.swing.JFrame;
import java.util.*;
import java.io.*;

import picture.*;
import picture.gui.*;


/**
 * PictureType-Unterklasse fr PGM-Dateien.
 *
 * @version 1.0
 * @author Stefan Rhrich
 */
public class PGMPictureType extends PictureType {


    static {
	PictureTypeFactory.registerPictureType(new PGMPictureType());
    }

    /**
     * Dummy zum Aufrufen des Static-Initializers ...
     */
    public static final void init() {
    }

    public String getExtension() {
	return "pgm";
    }

    public String getDescription() {
	return "Portable Graymap-Bild";
    }

    public DialogPane getSaveOptionsPane(JFrame f, Picture p) throws PictureException {
	return new SavePane(f, p);
    }

    public static final int ASCII = 1;
    public static final int RAW = 2;

    protected static final int VALUES_PER_LINE = 70;
    final static private String cp1 = 
	"# JIPP - Copyright (c) 2000 Manuel Kauers und Stefan Rhrich";
    final static private String cp2 = "#        Alle Rechte vorbehalten.";
    
    public Picture load(InputStream origin) throws PictureException, IOException {
	
	double offset = 0;
	double scale = 1;
	String[] comment = new String[1];

	InputStream in = new BufferedInputStream(origin);

	char t;
	Vector comments = new Vector();

	if((t = (char)in.read()) != 'P')
	    throw new PictureException("No PGM file + " + t + "!");

	int type = 0;

	switch(t = (char)in.read()) {
	case '2': type = ASCII; break;
	case '5': type = RAW;   break;
	default:
	    throw new PictureException("Unknown PGM type " + t + "!");
	}

	int width = readASCIINumber(skipWhitespacesGetComments(in, comments), in);
	int height = readASCIINumber(skipWhitespacesGetComments(in, comments), in);
	int mg = readASCIINumber(skipWhitespacesGetComments(in, comments), in);

	int pixelWidth = (int)Math.round(Math.log(mg)/(Math.log(2)*8.0)+0.5);

	boolean complex = false;

	if(getComment(comments, "isComplex") != null) {
	    complex = true;
	    width /= 2;
	}

	String is = getComment(comments, "intensityScale");

	if (is != null) {
	    StringTokenizer st = new StringTokenizer(is);
	    if(st.hasMoreTokens()) st.nextToken(); // skip name
	    if(st.hasMoreTokens()) offset = new Double(st.nextToken()).doubleValue();
	    if(st.hasMoreTokens()) scale = new Double(st.nextToken()).doubleValue();
	}

	double[][] re = new double[width][height];
	double[][] im = new double[width][height];

	for (int y = 0; y < height; y++)
	    for (int x = 0; x < width; x++)
		if (type == ASCII) {
		    re[x][y] = readASCIINumber(skipWhitespacesGetComments(in, comments), in)
			* scale + offset;
		    if (complex) 
			im[x][y] = readASCIINumber(skipWhitespacesGetComments(in, comments), in)
			    * scale + offset;
		} else {
		    re[x][y] = readRawValue(in, pixelWidth, scale, offset);
		    if (complex) 
			im[x][y] = readRawValue(in, pixelWidth, scale, offset);
		}

	in.close();

	removeComment(comments, "isComplex");
	removeComment(comments, "intensityScale");

	comment = vectorToStringArray(comments);

	Picture p;

	if (complex) {
	    p = new Picture(new Layer[]{new Layer(re), new Layer(im)});
	    p.setPresentation(PresentationFactory.presentationByName(p, "Komplexe Darstellung"));

	} else {
	    p = new Picture(new Layer[]{new Layer(re)});
	    p.setDefaultPresentation();
	}

	return p;
    }

    public void save(Picture p, PictureProperty pp, OutputStream origout) throws PictureException, IOException {
	BufferedOutputStream out = new BufferedOutputStream(origout);

	int t;

	long mg = 255;

	if(pp instanceof PGMPictureProperty) {
	    t = ((PGMPictureProperty)pp).getFormat();
	    mg = ((PGMPictureProperty)pp).getNumberOfGrayScales();
	}
	else
	    t = RAW;
      
	out.write('P');
	out.write(((t == RAW)?'5':'2'));
	out.write('\n');

	writeString(out, cp1);
	writeString(out, "\n");
	writeString(out, cp2);
	writeString(out, "\n");

	if(pp instanceof PGMPictureProperty) {
	    String[] w = ((PGMPictureProperty)pp).getComment();
	    for(int i = 0; i < w.length; i++)
		if(w[i] != null)
		    if(!cp1.equals(w[i]) && !cp2.equals(w[i]))
			writeString(out, "# " + w[i] + "\n");
	}

	double smin = p.getLayer(0).getMinimum();
	double smax = p.getLayer(0).getMaximum();

	double m = (smax-smin)/mg;


	writeString(out, "#intensityScale " + smin + " " + m + "\n");

	writeNumber(out, p.getWidth());
	out.write(' ');
	writeNumber(out, p.getHeight());
	out.write('\n');
	writeNumber(out, mg);
	out.write('\n');

	if(t == RAW)
	    writeRaw(out, p, smin, smax, mg);
	else
	    writeASCII(out, p, 20, smin, smax, mg);

	out.close();
    }
    

    /**
     * Skips whitespaces and comments (= "#" <String> "\n").
     * As length to skip is not known and file should be read only once, 
     * this method reads up to the next character. This character is returned.
     * @return first char which is not a whitespace or a comment char. 
     * @exception IOException on reading exception (e.g. EOFException).
     */ 
    protected static char skipWhitespacesGetComments (InputStream in, Vector comments) 
	throws IOException {
	
	boolean fertig = false;
	char c;

	do {
	    while(Character.isWhitespace(c = (char)in.read())) ;

	    if(c == '#') {
		String com = "";
		while((c = (char)in.read()) != '\n')
		    com += c;
		comments.addElement(com.trim());
	    }
	    else
		fertig = true;
	} while(!fertig);

	return c;
      
    } //skipWhitespacesAndComments()
    
    /**
     * Reads inputStream up to next whitespace or 
     * comment (= "#" <String> "\n") character. 
     * Builds up an int by concatenating
     * c with the following bytes from the inputStream.
     * @param c Character from input stream which was already read.
     * @param dataInputStream Stream to read on further.
     * @return generated number.
     * @exception IOException on reading exception (e.g. EOFException).
     * @exception PictureException if number contains non digit characters.
     */
    protected static int readASCIINumber(char c, InputStream in) 
	throws IOException, PictureException {
	
	StringBuffer t = new StringBuffer();
	char ct;
	
	while(!Character.isWhitespace(c) && c != '#') {
	    t.append(c);
	    c = (char)in.read();
	}

	return Integer.parseInt(t.toString());
      
    } //readASCIINumber()

    protected double readRawValue(InputStream in, int pw, double scale, double offset) throws IOException, PictureException {
	long v = 0;

	if(pw <= 0 || pw > 8)
	    throw new PictureException("PGM: Unknown pixel width!");

	if(pw >= 1)
	    v += ((in.read() & 0xff));
	if(pw >= 2)
	    v += ((in.read() & 0xff) << 8);
	if(pw >= 3)
	    v += ((in.read() & 0xff) << 16);
	if(pw >= 4)
	    v += ((in.read() & 0xff) << 24);
	if(pw >= 5)
	    v += ((in.read() & 0xff) << 32);
	if(pw >= 6)
	    v += ((in.read() & 0xff) << 40);
	if(pw >= 7)
	    v += ((in.read() & 0xff) << 48);
	if(pw == 8)
	    v += ((in.read() & 0xff) << 56);

	return v*scale+offset;
    }

    private String getComment(Vector comment, String search) {
	String t;

	for (Enumeration e = comment.elements() ; e.hasMoreElements() ;) {
	    if((t = (String)e.nextElement()).startsWith(search))
		return t;
	}
	return null;
    }

    private void removeComment(Vector comment, String search) {
	String t;

	for (Iterator e = comment.iterator() ; e.hasNext() ;) {
	    if(((String)e.next()).startsWith(search))
		e.remove();
	}
    }

    /**
     * Writes a string to specified OutputStream.
     *
     * @param out OutputStream to write to.
     * @param s String to be written.
     * @exception IOException Problems on writing to outputStream
     */
    protected static void writeString(OutputStream out, String s)
        throws IOException {
        out.write(s.getBytes());
    }

    protected void saveRawPixel(OutputStream out, long p, int pw) throws IOException {
	// keine breaks beabsichtigt!
	switch(pw) {
	case 8:
	    out.write((int)(p & 0xff));
	    p >>= 8;
	case 7:
	    out.write((int)(p & 0xff));
	    p >>= 8;
	case 6:
	    out.write((int)(p & 0xff));
	    p >>= 8;
	case 5:
	    out.write((int)(p & 0xff));
	    p >>= 8;
	case 4:
	    out.write((int)(p & 0xff));
	    p >>= 8;
	case 3:
	    out.write((int)(p & 0xff));
	    p >>= 8;
	case 2:
	    out.write((int)(p & 0xff));
	    p >>= 8;
	case 1:
	    out.write((int)(p & 0xff));
	}
    }

    protected void writeRaw(OutputStream out, Picture p, double min, double max, long mg)
	throws IOException, PictureException {
	int pw = (int)Math.round(Math.log(mg)/(Math.log(2)*8.0)+0.5);

	double[][] data = p.getLayer(0).data;

	for(int y = 0; y < p.getHeight(); y++)
	    for(int x = 0; x < p.getWidth(); x++)
		saveRawPixel(out, digitizePixel(data[x][y], mg, min, max), pw);
    } //writeRaw()

    protected void writeASCII(OutputStream out,	Picture p, final int valuesPerLine, double min, double max, long mg)
	throws IOException, PictureException {

	if(valuesPerLine > 70)
	    throw new PictureException("PGM: Too many values per line!");

	int lc = 0;

	double[][] data = p.getLayer(0).data;

	for(int y = 0; y < p.getHeight(); y++)
	    for(int x = 0; x < p.getWidth(); x++) {
		if((++lc % valuesPerLine) == 0)
		    out.write('\n');
		writeNumber(out, digitizePixel(data[x][y], mg, min, max));
		out.write(' ');
	    }
    } //writeASCII()

    private static final long digitizePixel(double d, long depth, double min, double max) {
        double m = depth / (max - min);

        if (d > max)
            return depth;
        else if (d < min)
            return 0;

        return (long)((d - min) * m);
    }
    /**
     * Writes a number as ASCII characters to specified OutputStream.
     *
     * @param out OutputStream to write to.
     * @param i int value to be written.
     * @exception IOException Problems on writing to outputStream
     */
    protected static void writeNumber(OutputStream out, long i)
        throws IOException {
        out.write(Long.toString(i).getBytes());
    }

    /**
     * Converts a vector of strings to String[].
     *
     * @param v Vector to convert.
     * @return generated String array.
     */
    protected static String[] vectorToStringArray(Vector v) {

        String[] s = new String[v.size()];

        for (int i = 0; i < v.size(); i++)
            s[i] = (String)v.elementAt(i);

        return s;
    }
}
