/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.filter;

import javax.swing.JFrame;
import picture.gui.*;
import picture.*;
import picture.math.*;

/**
 * Grauskalierung. Diese Transformation implementiert zwei Arten der Grauwertskalierung:
 * <p>
 * Zum einen kann eine lineare Grauwertskalierung durchgefuehrt werden, indem zwei
 * Parameter <tt>min</tt> und <tt>max</tt> angegeben werden. In diesem Fall werden alle
 * Farbwerte, die groesser als <tt>max</tt> bzw. kleiner als <tt>min</tt> sind, auf <tt>max</tt>
 * bzw <tt>min</tt> gesetzt.
 * <p>
 * Im allgemeineren Modus wird in Form einer Instanz der Klasse Term eine Skalierungsfunktion
 * in <i>x</i> angegeben, vermoege derer die Skalierung durchgefuehrt werden soll.
 * <p>
 * Die Grauwertskalierung bezieht sich stets auf die aktuelle Praesentation.
 *
 * @version 1.0
 */
public class GrayScale extends DefaultTransformation {

    public static final int LINEAR = 1;
    public static final int TERM = 2;
    
    private int mode;
    private Term t;
    private long depth;
    private double min;
    private double max;

    /**
     * Erzeugt eine Skalierung mit der angegebenen Skalierungsfunktion
     * @param t: Skalierungsfunktion
     * @param depth: Abtastrate der Funktion
     */
    public GrayScale(Term t, long depth) {setScaleFunction(t, depth);}

    /**
     * Erzeugt einen linearen Grauwertskalierung. Alles, was kleiner als
     * <tt>min</tt> ist, wird auf <tt>min</tt>, alles, was groesser als
     * <tt>max</tt> ist, wird auf <tt>max</tt> gesetzt. Diese Zahlen werden
     * absolut zwischen 0 und 1 angegeben.
     * @param min Untergrenze
     * @param max Obergrenze
     */
    public GrayScale(double min, double max) {setBounds(min, max);}

    /**
     * Gibt den aktuellen Modus zurueck. Dieser ist GrayScale.LINEAR oder
     * GrayScale.TERM
     */
    public int getMode() {return mode;}

    /**
     * Legt die Skalierungsfunktion fest
     * @param t: neue Skalierungsfunktion
     */
    public synchronized void setScaleFunction(Term t, long depth) {

	this.depth = depth; this.t = t; mode = TERM;
    }

    /**
     * Gibt die aktuelle Skalierungsfunktion zurueck.
     */
    public Term getScaleFunction() {return t;}

    /**
     * Legt neue Grenzen fest.
     * @param min neue Untergrenze
     * @param max neue Obergrenze
     */
    public synchronized void setBounds(double min, double max) {

	if (min > max) {
	    double m = min;
	    min = max; max = m;
	}

	this.min = min; this.max = max; mode = LINEAR;
    }

    /**
     * Gibt die Grenzen zurueck, wie sie mit <tt>getBounds(double, double)</tt> oder
     * im Konstruktor angegeben wurde.
     */
    public double [] getBounds() {

	return new double[]{min, max};
    }

    public int getNumberOfProgressSteps(Picture p) {

	return p.getNumberOfLayers() * p.getWidth();
    }

    /**
     * Diese Methode liefert ein neues Layer-Objekt mit der folgenden Eigenschaft:
     * <p>
     * 1. Fall: Diese Transformation ist linear. 
     * In der aktuellen Ansicht (Presentation) werden alle Grauwerte kleiner min auf
     * min gesetzt und alle groesseren Grauwerte als <tt>max</tt> auf <tt>max</tt> gesetzt. 
     * Die anderen Praesentationen werden unveraendert uebernommen bzw so angepasst, dass 
     * darin keine (bzw. moeglichst wenig) Aenderung sichtbar wird.
     * <p>
     * 2. Fall: Diese Transformation ist frei.
     * Skaliert die Grauwerte dieses Bildes anhand der uebergebenen Skalierungsfunktion.
     * Dazu wird diese Funktion auf dem Intervall [0, 1] ausgewertet: Habe 
     * <tt>(x, y)</tt> in der
     * aktuellen Praesentation den Wert <t>t</t>, so wird dieser Wert auf <tt>f(t)</tt>
     * gesetzt.
     * <p>
     * Diese Methode unterstuetzt die Funktion der ProgressHandler fuer die Verwendung
     * von Fortschrittsanzeigen (z.B. <tt>JProgressBar</tt>). Es werden 
     * <tt>width</tt> Events ausgeworfen
     * 
     * @param p Das Bild, das skaliert werden soll.
     * @return Skalierte Funktion von <tt>p</tt>
     */
    public synchronized Layer of(ProgressHandler ph, Layer p) throws PictureException {

	double min0 = p.getMinimum(); double max0 = p.getMaximum();
	double min1 = min; double max1 = max;

	Layer q = null;
	int width = p.getWidth(); int height = p.getHeight();
	double [][] data = new double[width][height];

	if (mode == LINEAR) {

	    double min = (max0 - min0) * min1 + min0;
	    double max = (max0 - min0) * max1 + min0;
	    
	    for (int x = 0; x < width; x++) {
		if (ph != null) ph.step();
		for (int y = 0; y < height; y++) {
		    double d = p.getValue(x, y);
		    if (d < min) d = min;
		    if (d > max) d = max;
		    data[x][y] = d;
		}
	    }
	    
	    q = createNewInstance(p, data);
		
	} else if (mode == TERM) {
		
	    String [] x0 = new String[] {"x"};
	    Complex [] z = new Complex[1];

	    if (depth < 0) depth = -depth;

	    double [] erg = new double[(int)depth + 1];
	    double min = Double.POSITIVE_INFINITY;
	    
	    try {
		for (int i = 0; i <= depth; i++) {
		    z[0] = new Complex(i * 1.0 / depth, 0);
		    erg[i] = t.einsetzen(x0, z).getRealPart();
		    if (erg[i] < min)
			min = erg[i];
		}
	    } catch (ArithmeticException e) {
		throw new PictureException(e.getMessage());
	    } catch (IllegalArgumentException e) {
		throw new PictureException("Syntaxfehler in der Skalierungsfunktion");
	    } catch (Exception e) {
		throw new PictureException("Ungltige Skalierungsfunktion");
	    }

	    if (min < 0) 
		for (int i = 0; i <= depth; i++)
		    erg[i] -= min;
	
	    min = p.getMinimum(); double max = p.getMaximum();
	    int index;

	    for (int x = 0; x < width; x++) {
		if (ph != null) ph.step();
		for (int y = 0; y < height; y++) {
		    index = (int) (depth * (p.getValue(x, y) - min) / (max - min));
		    data[x][y] = erg[index];
		}
	    }

	    q = createNewInstance(p, data);
	}

	return q;
    }

    /**
     * Liefert einen Optionendialog zurck.
     *
     * @param owner Parentframe
     * @param pic zu berbeitendes Bild
     * @return OptionenDialog
     */
    public DialogPane filterOptionPane(JFrame owner, Picture pic) 
	throws PictureException {

	return new ScalePane(owner, pic, this);
    }    
}
