/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.filter;

import picture.*;
import picture.math.*;

/**
 * Klasse zur Berechnung der Karhunen-Loeve-Transformation f&uuml;r ein mehrschichtiges Bild.
 * <p>
 * Sei <tt>p</tt> ein Picture mit <tt>n == p.getNumberOfLayers()</tt>. Dann dienen die 
 * Vektoren <tt>v</tt> mit
 * <pre>
 *  v[i] = p.getLayer(i).getDoubleValue(x, y);
 * </pre>
 * als Eingabe fuer die KLT, wobei <tt>(x, y)</tt> das gesamte Bild durchlauft. 
 * <p>
 * Ein KLT-Objekt wird initialisiert aus einem Picture <tt>p</tt> und dient dann zur
 * Basistransformation beliebiger Pictures (mit gleicher Ebenenzahl wie <tt>p</tt>) in
 * die KLT-Basis von <tt>p</tt> und wieder zur&uuml;ck.
 *
 * @version 0
 * @author Stefan R&ouml;hrich
 */
public class LayerKLT {

    private Picture p;
    private double[] mean;
    private int kltLength;
    private int numberOfVectors;
    KLT klt;

    /**
     * Erzeugt eine neue Instanz bez&uuml;glich des angegebenen Pictures.
     *
     * @param p: Picture mit den Eingabevektoren f&uuml;r die KLT.
     */
    public LayerKLT(Picture p) throws PictureException {
	this.p = p;

	kltLength = p.getNumberOfLayers();
	numberOfVectors = p.getWidth()*p.getHeight();
	double[][] vecs = new double[numberOfVectors][kltLength];
	mean = new double[kltLength];

	int i = 0;

	for(int x = 0; x < p.getWidth(); x++)
	    for(int y = 0; y < p.getHeight(); y++)
		vecs[i++] = p.getValue(x,y);

	//	klt = new KLT(vecs);
	klt = new KLT(KLT.normalize(vecs, mean));

	//	klt.printBase();
    }

    /** 
     * Gibt die bestimmten Eigenwerte in absteigender Reihenfolge aus.
     * Der <i>i</i>-te Eigenwert geh&ouml;rt dabei zum <i>i</i>-ten
     * Basisvektor der errechneten KLT-Basis. Insbesondere ist die
     * Anzahl der Eigenwerte gleich der Anzahl der Ebenen des ursp&uuml;nglichen
     * Bildes.
     *
     * @return Eigenwerte
     */
    public double[] getEigenValues() {
	return klt.getEigenValues();
    }

    private final double[] einheitsVektor(int i) {
	double[] out = new double[kltLength];

	out[i] = mean[i];

	return out;
    }
	
    /**
     * Gibt die bestimmten Eigenvektoren der KLT-Basis aus.  Die
     * Reihenfolge der Vektoren stimmt mit der Reihenfolge der
     * Eigenwerte &uuml;berein, die die Methode
     * <tt>getEigenValues()</tt> ausgibt:
     *
     * <pre>
     *  double[] v = getEigenVectors()[i]
     * </pre>
     * ist Eigenvektor zum Eigenwert <tt>getEigenValues()[i]</tt>.
     *
     * @return Eigenvektoren */
    public double[][] getEigenVectors() {
	double[][] evecs = new double[kltLength][kltLength];

	for(int i = 0; i < kltLength; i++)
	    evecs[i] = KLT.denormal(klt.inverseTransform(einheitsVektor(i)), mean);

	return evecs;
    }

    /**
     * F&uuml;hrt einen Basiswechsel durch. Es wird angenommen, da&szlig; das
     * angegebene Bild zur Standardbasis dargestellt ist. Hierzu wird eine
     * Darstellung der gleichen Daten zur KLT-Basis errechnet. Diese
     * Darstellung wird zur&uuml;ckgegeben. 
     * <p>
     * Diese Methode ist invers zu <tt>kltToStandard</tt>, d.h. es gilt
     * <tt>p == standardToKlt(kltToStandard(p)) == kltToStandard(standardToKlt(p))</tt>
     * f&uuml;r jedes Picture <tt>p</tt>.
     *
     * @see #kltToStandard
     * @param p: Bild zur Standardbasis
     * @return das gleiche Bild zur KLT-Basis
     */
    public Picture standardToKlt(Picture p) throws PictureException {
	return transform(p, false, null);
    }

    private Picture transform(Picture p, boolean inverse, double[] weights) 
	throws PictureException {

	if (p.getNumberOfLayers() != kltLength)
	    throw new PictureException("Falsche Bilddimension fr diese Basis!");

	double[][][] data = new double[p.getWidth()][p.getHeight()][kltLength];

	if (!inverse)
	    for(int x = 0; x < p.getWidth(); x++)
		for(int y = 0; y < p.getHeight(); y++)
		    data[x][y] = klt.transform(KLT.normal(p.getValue(x,y), mean));
		    //		    data[x][y] = klt.transform(p.getValue(x,y));
	else
	    if(weights == null)
		for(int x = 0; x < p.getWidth(); x++)
		    for(int y = 0; y < p.getHeight(); y++)
			data[x][y] = KLT.denormal(klt.inverseTransform(p.getValue(x,y)), mean);
		    //		    data[x][y] = klt.inverseTransform(p.getValue(x,y));
	    else
		for(int x = 0; x < p.getWidth(); x++)
		    for(int y = 0; y < p.getHeight(); y++)
			data[x][y] = KLT.denormal(klt.inverseTransform(p.getValue(x,y), weights), mean);

	double[][][] ldata = new double[kltLength][p.getWidth()][p.getHeight()];

	for(int lc = 0; lc < kltLength; lc++)
	    for(int x = 0; x < p.getWidth(); x++)
		for(int y = 0; y < p.getHeight(); y++)
		    ldata[lc][x][y] = data[x][y][lc];

	Layer[] layers = new Layer[kltLength];

        for(int lc = 0; lc < kltLength; lc++)
            layers[lc] = new Layer(ldata[lc]);

        return new Picture(p, layers);
    }

    /**
     * F&uuml;hrt einen Basiswechsel durch. Es wird angenommen, da&szlig; das
     * angegebene Bild zur KLT-Basis dargestellt ist. Hierzu wird die
     * korrespondierende Darstellung zur Standardbasis berechnet. Diese
     * Darstellung wird als Ergebnisbild zurueckgegeben.
     * <p>
     * Diese Methode ist invers zu <tt>standardToKlt</tt>, d.h. es gilt
     * <tt>p == standardToKlt(kltToStandard(p)) == kltToStandard(standardToKlt(p))</tt>
     * f&uuml;r jedes Picture <tt>p</tt>.
     *
     * @see #standardToKlt
     * @param p: Bild zur KLT-Basis
     * @return das gleiche Bild zur Standard-Basis
     */
    public Picture kltToStandard(Picture p) throws PictureException {
	return transform(p, true, null);
    }

    /**
     * Wie <tt>kltToStandard(p)</tt>, jedoch mit zus&auml;tzlicher Gewichtung der
     * Dimensionen.
     */
    public Picture kltToStandard(Picture p, double[] weights) throws PictureException {

	if (weights != null && weights.length != kltLength)
	    throw new PictureException("Ungltige Anzahl von Vorfaktoren");

	return transform(p, true, weights);
    }
}

