/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.io;

import java.util.*;
import java.io.*;

/**
 * Input stream which transforms its data via
 * Run-Length-Decoding.
 *
 * @version 0.1
 * @author Stefan Rhrich */
public class KnownLengthInputStream extends FilterInputStream {

    private ByteArrayInputStream buf = null;
 
    /**
     * Initialize and set input stream. */
    public KnownLengthInputStream(InputStream in) {
	super(in);
    }

    private int readLen() throws IOException {
	int t;
	if((t = in.read()) == -1)
	    return -1;
	int len = (t&0xff) << 24;
	if((t = in.read()) == -1)
	    throw new IOException("Invalid data (reading length 1).");
	len |= (t&0xff) << 16;
	if((t = in.read()) == -1)
	    throw new IOException("Invalid data (reading length 2).");
	len |= (t&0xff) << 8;
	if((t = in.read()) == -1)
	    throw new IOException("Invalid data (reading length 3).");
	len |= (t&0xff);
	byte[] b = new byte[len];

	int realsize = in.read(b);

	if(realsize != len)
	    throw new IOException("Invalid data: " + realsize + " instead of " + len + ".");

	buf = new ByteArrayInputStream(b);

	return len;
    }

    /**
     * Reads a byte from the underlying input stream. If the written
     * length is exceeded this is treated as EOF.
     *
     * @return tranformed byte or -1 for end of file
     * @exception IOException */
    public int read() throws IOException {
	if(buf == null && readLen() == -1)
	    return -1;

	return buf.read();
    }

    /**
     * Reads a byte array. If the written length is exceeded this is
     * treated as EOF.
     *
     * @param b where to write the bytes to
     * @param off starting offset
     * @param len how many bytes to read
     * @return number of bytes bytes actually read
     * @exception IOException */
    public int read(byte[] b, int off, int len) throws IOException {
	if(len == 0)
	    return 0;

	if(buf == null && readLen() == -1)
	    return -1;

	return buf.read(b, off, len);
    }

    /**
     * Return data from internal ByteArrayInputStream. */
    public void mark(int readlimit) {
	if(buf != null)
	    buf.mark(readlimit);
    }

    /**
     * Return data from internal ByteArrayInputStream. */
    public void reset() {
	if(buf != null)
	    buf.reset();
    }

    /**
     * Return data from internal ByteArrayInputStream. */
    public boolean markSupported() {
	if(buf != null)
	    return buf.markSupported();
	else
	    return false;
    }

    /**
     * Return data from internal ByteArrayInputStream. */
    public long skip(long n) throws IOException {
	if(buf != null)
	    return buf.skip(n);
	else
	    if(n > 0) {
		read();
		return buf.skip(n-1);
	    } else
		return 0;
    }

    /**
     * Return data from internal ByteArrayInputStream. */
    public int available() {
	if(buf != null)
	    return buf.available();
	else
	    return 0;
    }

    public static void main(String[] args) throws IOException {
	InputStream a = new KnownLengthInputStream(System.in);
        int t;
        while((t = a.read()) != -1)
            System.out.write(t);
        System.out.close();
    }
}
