/*
 * Copyright (c) 2000, 2001 Manuel Kauers, Stefan Rhrich
 * Alle Rechte vorbehalten.
 */

package picture.io;

/**
 * Simple MD5 message digest class, doesn't use or provide security
 * providers and other nice stuff in order to use it fast and easy
 * ;-).
 *
 * Java implementation of the RSA Data Security, Inc. MD5, derived
 * from the RSA Data Security, Inc. MD5 Message-Digest Algorithm. See
 * RFC 1321 for details.
 *
 * @author <A HREF="http://home.pages.de/~sr/">Stefan Roehrich</A> <A HREF="mailto:stefan@roehri.ch">stefan@roehri.ch</A> */
public final class MD5 {

    private static final String copyright = "Copyright(c) 1999 Stefan Rhrich, stefan@roehri.ch";

/*
 * The remaining code is the reference MD5 code (md5c.c) from rfc1321
 * portet to Java */
/* MD5C.C - RSA Data Security, Inc., MD5 message-digest algorithm
 */

/* Copyright (C) 1991-2, RSA Data Security, Inc. Created 1991. All
   rights reserved.

   License to copy and use this software is granted provided that it
   is identified as the "RSA Data Security, Inc. MD5 Message-Digest
   Algorithm" in all material mentioning or referencing this software
   or this function.

   License is also granted to make and use derivative works provided
   that such works are identified as "derived from the RSA Data
   Security, Inc. MD5 Message-Digest Algorithm" in all material
   mentioning or referencing the derived work.

   RSA Data Security, Inc. makes no representations concerning either
   the merchantability of this software or the suitability of this
   software for any particular purpose. It is provided "as is"
   without express or implied warranty of any kind.

   These notices must be retained in any copies of any part of this
   documentation and/or software.
 */

/* Constants for MD5Transform routine.
 */
    private static final int S11 = 7;
    private static final int S12 = 12;
    private static final int S13 = 17;
    private static final int S14 = 22;
    private static final int S21 = 5;
    private static final int S22 = 9;
    private static final int S23 = 14;
    private static final int S24 = 20;
    private static final int S31 = 4;
    private static final int S32 = 11;
    private static final int S33 = 16;
    private static final int S34 = 23;
    private static final int S41 = 6;
    private static final int S42 = 10;
    private static final int S43 = 15;
    private static final int S44 = 21;
    private static final byte[] PADDING =
    {
        (byte)0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    };

/* F, G, H and I are basic MD5 functions.
 */
    private static final int F(int x, int y, int z) {
	return (((x) & (y)) | ((~x) & (z)));
    }

    private static final int G(int x, int y, int z) {
	return (((x) & (z)) | ((y) & (~z)));
    }

    private static final int H(int x, int y, int z) {
	return ((x) ^ (y) ^ (z));
    }

    private static final int I(int x, int y, int z) {
	return ((y) ^ ((x) | (~z)));
    }

/* ROTATE_LEFT rotates x left n bits.
 */
    private static final int ROTATE_LEFT(int x, int n) {
	return (((x) << (n)) | ((x) >>> (32-(n))));
    }

/* FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
   Rotation is separate from addition to prevent recomputation.
 */
    private static final int FF(int a, int b, int c, int d, int x, int s, int ac) {
	a += F ((b), (c), (d)) + (x) + (ac);
	a = ROTATE_LEFT ((a), (s));
	a += (b);
	return a;
    }
    private static final int GG(int a, int b, int c, int d, int x, int s, int ac) {
	a += G ((b), (c), (d)) + (x) + (ac);
	a =  ROTATE_LEFT ((a), (s));
	a += (b);
	return a;
  }
    private static final int HH(int a, int b, int c, int d, int x, int s, int ac) {
	a += H ((b), (c), (d)) + (x) + (ac);
	a = ROTATE_LEFT ((a), (s));
	a += (b);
	return a;
  }
    private static final int II(int a, int b, int c, int d, int x, int s, int ac) {
	a += I ((b), (c), (d)) + (x) + (ac);
	a = ROTATE_LEFT ((a), (s));
	a += (b);
	return a;
  }

    /* Context: */
    /* magic initialization constants, state (ABCD) */
    private int[] state = {0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476};
    /* number of bits, modulo 2^64 (lsb first) */
    private int[] count = {0, 0};
    /* input buffer */
    private byte[] buffer = new byte[64];

    /* Currently unused */
    private final void initContext() {
        count[0] = count[1] = 0;
        /* Load magic initialization constants.
         */
        state[0] = 0x67452301;
        state[1] = 0xefcdab89;
        state[2] = 0x98badcfe;
        state[3] = 0x10325476;
    }

    private final void clearContext() {
        count[0] = count[1] = 0;
	state[0] = state[1] = state[2] = state[3] = 0;
	for(int i = 0; i < buffer.length; i++)
	    buffer[i] = 0;
    }

    /* Use only the static methods ... */
    private MD5() {
    }
	
/* MD5 block update operation. Continues an MD5 message-digest
   operation, processing another message block, and updating the
   context.
 */
    private final void md5Update(final byte[] input, int inputLen) {
        int i, index, partLen;

        /* Compute number of bytes mod 64 */
        index = (int) ((count[0] >>> 3) & 0x3F);

        /* Update number of bits */
        if ((count[0] += (inputLen << 3)) < (inputLen << 3))
	    count[1]++;
        count[1] += (inputLen >>> 29);

        partLen = 64 - index;

        /* Transform as many times as possible.
         */
        if (inputLen >= partLen) {
	    System.arraycopy(input, 0, buffer, index, partLen);
	    md5Transform(buffer);
		
	    byte[] t = new byte[input.length];
	    for(i = partLen; i + 63 < inputLen; i += 64) {
		System.arraycopy(input, i, t, 0, input.length-i);
		md5Transform(t);
	    }

	    index = 0;
        } else
	    i = 0;

        /* Buffer remaining input */
        System.arraycopy(input, i, buffer, index, inputLen - i);
    }

/* MD5 finalization. Ends an MD5 message-digest operation, writing the
   message digest and zeroizing the context. */
    private final void md5Final(byte[] digest) {
        byte[] bits = new byte[8];
        int index, padLen;

        /* Save number of bits */
        encode(bits, count, 8);

        /* Pad out to 56 mod 64.
         */
        index = (int) ((count[0] >>> 3) & 0x3f);
        padLen = (index < 56) ? (56 - index) : (120 - index);
        md5Update(PADDING, padLen);

        /* Append length (before padding) */
        md5Update(bits, 8);

        /* Store state in digest */
        encode(digest, state, 16);

        /* Zeroize sensitive information.
         */
	clearContext();
    }

/* MD5 basic transformation. Transforms state based on block.
 */
    private final void md5Transform(final byte[] block) {
        int a = state[0], b = state[1], c = state[2], d = state[3];
	int[] x = new int[16];

        decode(x, block, 64);

        /* Round 1 */
        a = FF(a, b, c, d, x[0], S11, 0xd76aa478);  /* 1 */
        d = FF(d, a, b, c, x[1], S12, 0xe8c7b756);  /* 2 */
        c = FF(c, d, a, b, x[2], S13, 0x242070db);  /* 3 */
        b = FF(b, c, d, a, x[3], S14, 0xc1bdceee);  /* 4 */
        a = FF(a, b, c, d, x[4], S11, 0xf57c0faf);  /* 5 */
        d = FF(d, a, b, c, x[5], S12, 0x4787c62a);  /* 6 */
        c = FF(c, d, a, b, x[6], S13, 0xa8304613);  /* 7 */
        b = FF(b, c, d, a, x[7], S14, 0xfd469501);  /* 8 */
        a = FF(a, b, c, d, x[8], S11, 0x698098d8);  /* 9 */
        d = FF(d, a, b, c, x[9], S12, 0x8b44f7af);  /* 10 */
        c = FF(c, d, a, b, x[10], S13, 0xffff5bb1);         /* 11 */
        b = FF(b, c, d, a, x[11], S14, 0x895cd7be);         /* 12 */
        a = FF(a, b, c, d, x[12], S11, 0x6b901122);         /* 13 */
        d = FF(d, a, b, c, x[13], S12, 0xfd987193);         /* 14 */
        c = FF(c, d, a, b, x[14], S13, 0xa679438e);         /* 15 */
        b = FF(b, c, d, a, x[15], S14, 0x49b40821);         /* 16 */

        /* Round 2 */
        a = GG(a, b, c, d, x[1], S21, 0xf61e2562);  /* 17 */
        d = GG(d, a, b, c, x[6], S22, 0xc040b340);  /* 18 */
        c = GG(c, d, a, b, x[11], S23, 0x265e5a51);         /* 19 */
        b = GG(b, c, d, a, x[0], S24, 0xe9b6c7aa);  /* 20 */
        a = GG(a, b, c, d, x[5], S21, 0xd62f105d);  /* 21 */
        d = GG(d, a, b, c, x[10], S22, 0x2441453);  /* 22 */
        c = GG(c, d, a, b, x[15], S23, 0xd8a1e681);         /* 23 */
        b = GG(b, c, d, a, x[4], S24, 0xe7d3fbc8);  /* 24 */
        a = GG(a, b, c, d, x[9], S21, 0x21e1cde6);  /* 25 */
        d = GG(d, a, b, c, x[14], S22, 0xc33707d6);         /* 26 */
        c = GG(c, d, a, b, x[3], S23, 0xf4d50d87);  /* 27 */
        b = GG(b, c, d, a, x[8], S24, 0x455a14ed);  /* 28 */
        a = GG(a, b, c, d, x[13], S21, 0xa9e3e905);         /* 29 */
        d = GG(d, a, b, c, x[2], S22, 0xfcefa3f8);  /* 30 */
        c = GG(c, d, a, b, x[7], S23, 0x676f02d9);  /* 31 */
        b = GG(b, c, d, a, x[12], S24, 0x8d2a4c8a);         /* 32 */

        /* Round 3 */
        a = HH(a, b, c, d, x[5], S31, 0xfffa3942);  /* 33 */
        d = HH(d, a, b, c, x[8], S32, 0x8771f681);  /* 34 */
        c = HH(c, d, a, b, x[11], S33, 0x6d9d6122);         /* 35 */
        b = HH(b, c, d, a, x[14], S34, 0xfde5380c);         /* 36 */
        a = HH(a, b, c, d, x[1], S31, 0xa4beea44);  /* 37 */
        d = HH(d, a, b, c, x[4], S32, 0x4bdecfa9);  /* 38 */
        c = HH(c, d, a, b, x[7], S33, 0xf6bb4b60);  /* 39 */
        b = HH(b, c, d, a, x[10], S34, 0xbebfbc70);         /* 40 */
        a = HH(a, b, c, d, x[13], S31, 0x289b7ec6);         /* 41 */
        d = HH(d, a, b, c, x[0], S32, 0xeaa127fa);  /* 42 */
        c = HH(c, d, a, b, x[3], S33, 0xd4ef3085);  /* 43 */
        b = HH(b, c, d, a, x[6], S34, 0x4881d05);   /* 44 */
        a = HH(a, b, c, d, x[9], S31, 0xd9d4d039);  /* 45 */
        d = HH(d, a, b, c, x[12], S32, 0xe6db99e5);         /* 46 */
        c = HH(c, d, a, b, x[15], S33, 0x1fa27cf8);         /* 47 */
        b = HH(b, c, d, a, x[2], S34, 0xc4ac5665);  /* 48 */

        /* Round 4 */
        a = II(a, b, c, d, x[0], S41, 0xf4292244);  /* 49 */
        d = II(d, a, b, c, x[7], S42, 0x432aff97);  /* 50 */
        c = II(c, d, a, b, x[14], S43, 0xab9423a7);         /* 51 */
        b = II(b, c, d, a, x[5], S44, 0xfc93a039);  /* 52 */
        a = II(a, b, c, d, x[12], S41, 0x655b59c3);         /* 53 */
        d = II(d, a, b, c, x[3], S42, 0x8f0ccc92);  /* 54 */
        c = II(c, d, a, b, x[10], S43, 0xffeff47d);         /* 55 */
        b = II(b, c, d, a, x[1], S44, 0x85845dd1);  /* 56 */
        a = II(a, b, c, d, x[8], S41, 0x6fa87e4f);  /* 57 */
        d = II(d, a, b, c, x[15], S42, 0xfe2ce6e0);         /* 58 */
        c = II(c, d, a, b, x[6], S43, 0xa3014314);  /* 59 */
        b = II(b, c, d, a, x[13], S44, 0x4e0811a1);         /* 60 */
        a = II(a, b, c, d, x[4], S41, 0xf7537e82);  /* 61 */
        d = II(d, a, b, c, x[11], S42, 0xbd3af235);         /* 62 */
        c = II(c, d, a, b, x[2], S43, 0x2ad7d2bb);  /* 63 */
        b = II(b, c, d, a, x[9], S44, 0xeb86d391);  /* 64 */

        state[0] += a;
        state[1] += b;
        state[2] += c;
        state[3] += d;

        /* Zeroize sensitive information. */
	for(int i = 0; i < x.length; i++)
	    x[i] = 0;
    }

/* Encodes input (UINT4) into output (unsigned char). Assumes len is
   a multiple of 4.
 */
    private static final void encode(byte[] output, int[] input, int len) {
        int i, j;

        for (i = 0, j = 0; j < len; i++, j += 4) {
	    output[j] = (byte) (input[i] & 0xff);
	    output[j + 1] = (byte) ((input[i] >>> 8) & 0xff);
	    output[j + 2] = (byte) ((input[i] >>> 16) & 0xff);
	    output[j + 3] = (byte) ((input[i] >>> 24) & 0xff);
        }
    }

/* Decodes input (unsigned char) into output (UINT4). Assumes len is
   a multiple of 4.
 */
    private static final void decode(int[] output, final byte[] input, int len) {
        int i, j;

        for (i = 0, j = 0; j < len; i++, j += 4)
	    output[i] = (ByteUtils.unsignedByte2Int(input[j])) | ((ByteUtils.unsignedByte2Int(input[j + 1])) << 8) |
		((ByteUtils.unsignedByte2Int(input[j + 2])) << 16) | ((ByteUtils.unsignedByte2Int(input[j + 3])) << 24);
    }

    /**
     * Returns the raw bits of the MD5 hash function of the input
     * data.
     *
     * @param data Data
     *
     * @return md5 checksum as raw byte array
     *
     * @see #md5(byte[] data) */
    public static final byte[] md5raw(byte[] data) {
        byte[] digest = new byte[16];
	MD5 md5work = new MD5();

        md5work.md5Update(data, data.length);
        md5work.md5Final(digest);
	return digest;
    }

    /**
     * Returns the result of the MD5 hash function of the input data
     * as hex string.
     *
     * @param data Data
     *
     * @return md5 checksum as 32 characters long hex string
     *
     * @see #md5raw(byte[] data) */
    public static final String md5(byte[] data) {
        StringBuffer md5str = new StringBuffer(32);
	byte[] md5bits = md5raw(data);

	for(int i = 0; i < 16; i++) {
	    String t = Integer.toHexString(ByteUtils.unsignedByte2Int(md5bits[i]));
	    if(t.length() == 1)
		md5str.append('0');
	    md5str.append(t);
	}

	return md5str.toString();
    }

    /**
     * Returns the result of the MD5 hash function of the input data
     * as hex string. The data string will be converted to a byte
     * array, so beware of character encodings!
     *
     * @param data Data as string
     *
     * @return md5 checksum as 32 characters long hex string
     *
     * @see #md5raw(byte[] data) */
    public static final String md5(String data) {
	return md5(data.getBytes());
    }
}
